function [S,compositeIndices,pbStack,imgNdx,Slabels,pb,pbTheta] = DataDrivenSegmentation(img,G,pb,pbTheta,DB,HOMEGABOR)
% DataDrivenSegmentation.m
%
% Main function for NIPS 2009 paper.
% 
% Input:
% img - Input image
% G - Gist descriptors for image stack
% pb - PB response for input image (optional)
% pbTheta - PB theta orientations for input image (optional)
% DB - LabelMe DB structure for image stack
% HOMEGABOR - Path to pre-computed Gabor & color features
%
% Outputs:
% S - Segmentation map
% compositeIndices - Each row contains the NN segment indices (indexing
%                    into the input DB structure)
% pbStack - Data-driven boundary output
% imgNdx - Global gist NN image indices
% Slabels - Latent aspects for segments
% pb - Output of PB
% pbTheta - Orientation at each PB output

% Parameters:
imageSize = [375 500];
K = 5000; % Number of gist nearest neighbors
allGaborCellSize = 8; % Gabor block size
radiusWing = 24; % Butterfly wing radius (in pixels)
Nk = 1000; % Number of non-zero elements in local query binary vector
radiusLocal = 24; % Radius of local queries
grid_spacing = 16; % Spacing between local query sampling points
Ngrid = [16 12]; % Number of cells for segment queries
Nclusters = 5; % Number of latent aspects
alpha = 0.1; % GraphCuts Potts parameter
beta = 0.25; % GraphCuts boundary parameter
outlierThresh = -0.25; % Unary cost for GraphCuts outlier state
% $$$ Ncomp = 9; % Number of composites to return

% Gist parameters:
param.orientationsPerScale = [8 8 8 8];
param.fc_prefilt = 4;
param.imageSize = imageSize;
param.G = createGabor(param.orientationsPerScale,param.imageSize);

% Resize image:
img = imresize(img,imageSize,'bicubic');

% Compute gist NN:
gQuery = getfield(runGistIm(img),'D');
d = distMatL1(double(gQuery),double(G))/length(gQuery);
[lastDists,imgNdx] = sort(d);
lastDists = lastDists(1:K);
imgNdx = imgNdx(1:K);

% Get allGabor:
% $$$ HOMEGABOR = '/csail/d1/csail/streets100K_2/allGaborColor';
allGabor = getAllGabor_v2(DB,imgNdx,HOMEGABOR);
% $$$ GABOR_DIR = '~/work/FlickrExpts/disc/Cascade/allGaborColor';
% $$$ allGabor = getAllGabor(imgNdx,GABOR_DIR);

% Get Gabor responses for input image:
queryGabor = single(gridResponse(computeGabor(img,param),allGaborCellSize));
queryGabor = addColorFeatures(img,queryGabor);

% Compute PB:
if nargin < 4 || isempty(pb)
  [pb,pbTheta] = pbCGTG(double(img)/255);
end

%%% Step 1: Data-driven boundary detection %%%

% Compute butterflies over PB responses:
[Rleft,Rright,Rndx] = denseButterflyPB(queryGabor,allGabor,allGaborCellSize,pb,pbTheta,radiusWing);
pbStack = ranks2PBstack(Rleft,Rright,Rndx,pb);

%%% Step 2: Data-driven region cues %%%

% Get sample points:
[X,Y] = meshgrid(radiusLocal+1:grid_spacing:imageSize(2)-radiusLocal,radiusLocal+1:grid_spacing:imageSize(1)-radiusLocal);

% Do local queries:
Np = length(X(:));
R = zeros(size(allGabor,4),Np);
for i = 1:Np
  display(sprintf('%d out of %d',i,Np));
  c = [X(i) Y(i)];
  [xx,yy] = meshgrid(1:imageSize(2),1:imageSize(1));
  w = exp(-0.5*((xx-c(1)).^2+(yy-c(2)).^2)/radiusLocal^2);
  wDown = single(gridResponse(w,allGaborCellSize));
  d = queryFlickrWeighted(queryGabor,allGabor,Ngrid,wDown);
  [v,n] = sort(d);
  R(n,i) = [1:size(allGabor,4)];
end

% Run kmeans on top K-ranked images:
binaryVectors = double(R<=Nk);
[nn,ctrs] = kmeans(binaryVectors',Nclusters);

% Compute distance to cluster centers (unary terms):
unaryTerms = zeros(size(X,1),size(X,2),Nclusters);
for i = 1:Nclusters
  d = sum(binaryVectors.*repmat(ctrs(i,:)',1,Np),1);
  unaryTerms(:,:,i) = reshape(d,size(X,1),size(X,2));
end

%%% Step 3: GraphCuts %%%

% Run GraphCuts:
labels = rankGraphCut(-unaryTerms/Nk,X,Y,pbStack,pbTheta,alpha,beta,outlierThresh);

% Get segmentation map:
S = zeros(size(labels));
Slabels = [];
tt = 0;
for k = 1:Nclusters+1
  ll = bwlabel(labels==k);
  for ss = 1:max(ll(:))
    mask = (ll==ss);
    tt = tt+1;
    S(mask) = tt;
    Slabels(tt) = k;
  end
end


% Get image indices for composites:
compositeIndices = zeros(max(S(:)),size(allGabor,4));
for k = 1:max(S(:))
  w = (S==k);
  wDown = single(gridResponse(w,allGaborCellSize));
  d = queryFlickrWeighted(queryGabor,allGabor,Ngrid,wDown);
  [v,n] = sort(d);
  compositeIndices(k,:) = imgNdx(n);
end


% $$$ % Get composites:
% $$$ imgComposite = [];
% $$$ if nargin >= 6
% $$$   imgComposite = zeros(imageSize(1),imageSize(2),3,Ncomp,'uint8');
% $$$   for k = 1:max(S(:))
% $$$     w = (S==k);
% $$$     wDown = single(gridResponse(w,allGaborCellSize));
% $$$     d = queryFlickrWeighted(queryGabor,allGabor,Ngrid,wDown);
% $$$     [v,n] = sort(d);
% $$$     stack = getImageStack(DB,HOMEIMAGES,imgNdx(n),imageSize,Ncomp-1);
% $$$     imgComposite(repmat(w,[1 1 3 Ncomp])) = stack(repmat(w,[1 1 3 Ncomp]));
% $$$   end
% $$$   imgComposite(:,:,:,1) = img;
% $$$   for k = 1:Ncomp
% $$$     timg = overlaySegmentation(imgComposite(:,:,:,k),S,[0 0 0],12);
% $$$     imgComposite(:,:,:,k) = overlaySegmentation(timg,S,[255 255 0],4);
% $$$   end
% $$$ % $$$   imgComposite(:,:,:,1) = overlaySegmentation(img,S);
% $$$ % $$$   imgOverlay = overlaySegmentation(img,S);
% $$$ % $$$   imgComposite(:,:,:,1) = [];
% $$$ end
